"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const oracledb_1 = __importDefault(require("oracledb"));
const defaultOptions = {
    checkExpirationInterval: 900000,
    expiration: 86400000,
    createDatabaseTable: true,
    schema: {
        tableName: "sessions",
        columnNames: {
            /* eslint-disable @typescript-eslint/camelcase */
            session_id: "session_id",
            expires: "expires",
            data: "attributes",
        },
    },
};
function isObject(a) {
    const type = typeof a;
    return type === "function" || (type === "object" && !!a);
}
function deepcopy(a, b) {
    if (!isObject(a)) {
        return a;
    }
    for (const key in a) {
        if (b[key] === undefined) {
            b[key] = a[key];
        }
        b[key] = deepcopy(a[key], b[key]);
    }
    return b;
}
// because we export declaration we must explicitly set SessionInstance["Store"]
// type to avoid private type complain
const store = (session) => {
    class OracleSessionStore extends session.Store {
        constructor(options) {
            super(options);
            this._expirationInterval = 0;
            // doing deep copy to options
            this.options = deepcopy(defaultOptions, options);
            // create oracle pool
            this.pool = this.createNewPool();
            if (this.options.createDatabaseTable) {
                this.createDatabaseTable().then(() => {
                    this.setExpirationInterval();
                });
            }
            else {
                this.setExpirationInterval();
            }
        }
        createNewPool() {
            return oracledb_1.default.createPool(this.options).catch(error => {
                console.error(error, "express-oracle-session-ts: db pool is failed to initialized, app will exit with code 1");
                process.exit(1);
            });
        }
        getCon(activePool, tryCount) {
            return activePool.then(mypool => mypool.getConnection().catch(error => {
                console.error(error, `cannot get connection number of tries left: ${tryCount}`);
                // close the pool
                mypool.close(10);
                // if we have tried enough times
                if (tryCount === 0) {
                    console.log("Give up trying new pool, EOST is exiting with code 1");
                    return process.exit(1);
                }
                // create new pool promise
                this.pool = this.createNewPool();
                return this.getCon(this.pool, tryCount - 1);
            }));
        }
        getOraConn() {
            return this.getCon(this.pool, 3);
        }
        createDatabaseTable() {
            return __awaiter(this, void 0, void 0, function* () {
                const sql = `CREATE TABLE ${this.options.schema.tableName} (
${this.options.schema.columnNames.session_id} VARCHAR(128)  NOT NULL,
${this.options.schema.columnNames.expires} NUMBER(11) NOT NULL,
${this.options.schema.columnNames.data} CLOB,
PRIMARY KEY (${this.options.schema.columnNames.session_id})
)`;
                let conn;
                try {
                    conn = yield this.getOraConn();
                    yield conn.execute(sql);
                }
                catch (error) {
                    if (error.message.substring(0, 9) === "ORA-00955") {
                        console.log("EOST: Table already exists");
                        return;
                    }
                    console.error("Failed to create sessions database table.");
                    console.error(error);
                }
                finally {
                    if (conn) {
                        conn.close();
                    }
                }
            });
        }
        get(sid, cb) {
            const sql = `SELECT ${this.options.schema.columnNames.data} AS data FROM ${this.options.schema.tableName} WHERE ${this.options.schema.columnNames.session_id} = :sessionid AND ROWNUM = 1`;
            const params = { sessionid: sid };
            this.getOraConn().then(conn => {
                conn
                    .execute(sql, params, { fetchInfo: { DATA: { type: oracledb_1.default.STRING } } } // Fetch as a String instead of a Stream
                )
                    .then(result => {
                    let session = null;
                    try {
                        session = (result === null || result === void 0 ? void 0 : result.rows) && (result === null || result === void 0 ? void 0 : result.rows[0]) ? JSON.parse(result === null || result === void 0 ? void 0 : result.rows[0][0]) : null;
                    }
                    catch (error) {
                        console.error("EOTS: get error", error);
                        return cb(new Error("Failed to parse data for session: " + sid));
                    }
                    cb(null, session);
                })
                    .catch(error => {
                    return cb(new Error(error));
                })
                    .finally(() => conn.close());
            });
        }
        set(sid, data, cb) {
            let expires;
            if (data.cookie) {
                if (data.cookie.expires) {
                    expires = data.cookie.expires;
                }
            }
            if (!expires) {
                expires = Date.now() + this.options.expiration;
            }
            if (!(expires instanceof Date)) {
                expires = new Date(expires);
            }
            // Use whole seconds here; not milliseconds.
            expires = Math.round(expires.getTime() / 1000);
            const jsondata = JSON.stringify(data);
            const sql = `MERGE INTO ${this.options.schema.tableName} trg
USING (SELECT :sessionid as ${this.options.schema.columnNames.session_id} FROM DUAL) src
ON (trg.${this.options.schema.columnNames.session_id} = src.${this.options.schema.columnNames.session_id})
WHEN MATCHED THEN UPDATE
SET trg.${this.options.schema.columnNames.expires} = :expires,
trg.${this.options.schema.columnNames.data} = :attributes
WHEN NOT MATCHED THEN INSERT (${this.options.schema.columnNames.session_id}, ${this.options.schema.columnNames.expires}, ${this.options.schema.columnNames.data})
VALUES (:sessionid,:expires,:attributes)`;
            const params = {
                sessionid: sid,
                expires: expires,
                attributes: jsondata,
            };
            this.getOraConn().then(conn => {
                conn
                    .execute(sql, params, { autoCommit: true })
                    .then(() => {
                    cb && cb();
                })
                    .catch(error => {
                    return cb && cb(error);
                })
                    .finally(() => conn.close());
            });
        }
        touch(sid, data, cb) {
            let expires;
            if (data.cookie) {
                if (data.cookie.expires) {
                    expires = data.cookie.expires;
                }
            }
            if (!expires) {
                expires = Date.now() + this.options.expiration;
            }
            if (!(expires instanceof Date)) {
                expires = new Date(expires);
            }
            // Use whole seconds here; not milliseconds.
            expires = Math.round(expires.getTime() / 1000);
            const sql = `UPDATE ${this.options.schema.tableName} SET ${this.options.schema.columnNames.expires} = :expires WHERE ${this.options.schema.columnNames.session_id} = :sessionid`;
            const params = {
                expires: expires,
                sessionid: sid,
            };
            this.getOraConn().then(conn => {
                conn
                    .execute(sql, params, { autoCommit: true })
                    .then(() => {
                    cb && cb();
                })
                    .catch(error => {
                    return cb && cb(error);
                })
                    .finally(() => conn.close());
            });
        }
        destroy(sid, cb) {
            const sql = `DELETE FROM ${this.options.schema.tableName} WHERE ${this.options.schema.columnNames.session_id} = :sessionid`;
            const params = {
                sessionid: sid,
            };
            this.getOraConn().then(conn => {
                conn
                    .execute(sql, params, { autoCommit: true })
                    .then(() => {
                    cb && cb();
                })
                    .catch(error => {
                    cb && cb(error);
                })
                    .finally(() => conn.close());
            });
        }
        length(cb) {
            const sql = `SELECT COUNT(*) FROM ${this.options.schema.tableName}`;
            this.getOraConn().then(conn => {
                conn
                    .execute(sql)
                    .then(result => {
                    const count = result.rows && result.rows[0] ? result.rows[0][0] : 0;
                    cb(null, count);
                })
                    .catch(error => {
                    cb(error);
                })
                    .finally(() => conn.close());
            });
        }
        clear(cb) {
            const sql = `DELETE FROM ${this.options.schema.tableName}`;
            this.getOraConn().then(conn => {
                conn
                    .execute(sql)
                    .then(() => {
                    cb && cb();
                })
                    .catch(error => {
                    cb && cb(error);
                })
                    .finally(() => conn.close());
            });
        }
        clearExpiredSessions(cb) {
            const sql = `DELETE FROM ${this.options.schema.tableName} WHERE ${this.options.schema.columnNames.expires} < :expir`;
            const params = {
                expir: Math.round(Date.now() / 1000),
            };
            this.getOraConn().then(conn => {
                conn
                    .execute(sql, params, { autoCommit: true })
                    .then(() => {
                    cb && cb();
                })
                    .catch(error => {
                    cb && cb(error);
                })
                    .finally(() => conn.close());
            });
        }
        setExpirationInterval(interval) {
            interval || (interval = this.options.checkExpirationInterval);
            setInterval(() => this.clearExpiredSessions, interval);
        }
    }
    return OracleSessionStore;
};
exports.default = store;
